{
    "CBAR": {
        "EID": int,
        "G0": None,
        "G0/X1": float,
        "GA": int,
        "GB": int,
        "PA": None,
        "PB": None,
        "PID": int,
        "W1A": float,
        "W1B": float,
        "W2A": float,
        "W2B": float,
        "W3A": float,
        "W3B": float,
        "X1": float,
        "X2": float,
        "X3": float,
    },
    "CBUSH": {
        "CID": int,
        "EID": int,
        "GA": int,
        "GB": int,
        "GO/X1": float,
        "OCID": int,
        "PID": int,
        "S": float,
        "S1": None,
        "S2": None,
        "S3": None,
        "X2": float,
        "X3": float,
    },
    "CONM2": {
        "CID": int,
        "EID": int,
        "G": int,
        "I11": float,
        "I21": float,
        "I22": float,
        "I31": float,
        "I32": float,
        "I33": float,
        "M": float,
        "X1": float,
        "X2": float,
        "X3": float,
    },
    "CORD2C": {
        "A1": float,
        "A2": float,
        "A3": float,
        "B1": float,
        "B2": float,
        "B3": float,
        "C1": float,
        "C2": float,
        "C3": float,
        "CID": int,
        "RID": int,
    },
    "CORD2R": {
        "A1": float,
        "A2": float,
        "A3": float,
        "B1": float,
        "B2": float,
        "B3": float,
        "C1": float,
        "C2": float,
        "C3": float,
        "CID": int,
        "RID": int,
    },
    "CORD2S": {
        "A1": float,
        "A2": float,
        "A3": float,
        "B1": float,
        "B2": float,
        "B3": float,
        "C1": float,
        "C2": float,
        "C3": float,
        "CID": int,
        "RID": int,
    },
    "CQUAD4": {
        "EID": int,
        "G1": int,
        "G2": int,
        "G3": int,
        "G4": int,
        "MCID": int,
        "PID": int,
        "T1": None,
        "T2": None,
        "T3": None,
        "T4": None,
        "TFLAG": None,
        "THETA": float,
        "THETA/MCID": int,
        "ZOFFS": None,
    },
    "CROD": {"EID": int, "G1": int, "G2": int, "PID": int, "SM": None},
    "CTRIA3": {
        "EID": int,
        "G1": int,
        "G2": int,
        "G3": int,
        "MCID": int,
        "PID": int,
        "T1": None,
        "T2": None,
        "T3": None,
        "TFLAG": None,
        "THETA": float,
        "THETA/MCID": float,
        "ZOFFS": None,
    },
    "FORCE": {
        "CID": int,
        "F": float,
        "G": int,
        "N1": float,
        "N2": float,
        "N3": float,
        "SID": int,
    },
    "GRAV": {
        "A": float,
        "CID": int,
        "MB": int,
        "N1": float,
        "N2": float,
        "N3": float,
        "SID": int,
    },
    "GRID": {
        "CD": int,
        "CP": int,
        "ID": int,
        "PS": int,
        "SEID": None,
        "X1": float,
        "X2": float,
        "X3": float,
    },
    "LOAD": {"S": float, "SID": int, "load_factorsID": int},
    "MAT1": {
        "A": float,
        "E": float,
        "G": float,
        "GE": float,
        "MCSID": None,
        "MID": int,
        "NU": float,
        "RHO": float,
        "SC": None,
        "SS": None,
        "ST": None,
        "TREF": float,
    },
    "MAT8": {
        "A1": float,
        "A2": float,
        "E1": float,
        "E2": float,
        "F12": None,
        "G12": float,
        "G1Z": float,
        "G2Z": float,
        "GE": None,
        "MID": int,
        "NU12": float,
        "RHO": float,
        "S": None,
        "STRN": None,
        "TREF": float,
        "Xc": None,
        "Xt": None,
        "Yc": None,
        "Yt": None,
    },
    "PBAR": {
        "A": float,
        "C1": float,
        "C2": float,
        "D1": float,
        "D2": float,
        "E1": float,
        "E2": float,
        "F1": float,
        "F2": float,
        "I1": float,
        "I12": float,
        "I2": float,
        "J": float,
        "K1": float,
        "K2": float,
        "MID": int,
        "NSM": float,
        "PID": int,
    },
    "PBUSH": {
        '"B"': str,
        '"GE"': str,
        '"K"': str,
        '"RCV"': str,
        "B1": None,
        "B2": None,
        "B3": None,
        "B4": None,
        "B5": None,
        "B6": None,
        "EA": None,
        "ET": None,
        "GE1": None,
        "GE2": None,
        "GE3": None,
        "GE4": None,
        "GE5": None,
        "GE6": None,
        "K1": float,
        "K2": float,
        "K3": float,
        "K4": float,
        "K5": float,
        "K6": float,
        "PID": int,
        "SA": None,
        "ST": None,
    },
    "PCOMP": {
        "FT": None,
        "GE": float,
        "LAM": None,
        "NSM": float,
        "PID": int,
        "SB": None,
        "TREF": None,
        "Z0": None,
        "pcomp_layupID": int,
    },
    "PLOAD4": {
        "CID": None,
        "EID": int,
        "G1": None,
        "G1orG4": None,
        "N1": None,
        "N2": None,
        "N3": None,
        "P1": float,
        "P2": None,
        "P3": None,
        "P4": None,
        "SID": int,
    },
    "PROD": {"A": float, "C": float, "J": float, "MID": int, "NSM": float, "PID": int},
    "PSHELL": {
        "12I/T**3": float,
        "MID1": int,
        "MID2": int,
        "MID3": int,
        "MID4": None,
        "NSM": float,
        "PID": int,
        "T": float,
        "TS/T": float,
        "Z1": None,
        "Z2": None,
    },
    "RBE2": {"ALPHA": float, "CM": int, "EID": int, "GN": int, "rbe2_gidsetID": int},
    "RBE3": {
        "ALPHA": float,
        "EID": int,
        "REFC": int,
        "REFGRID": int,
        "rbe3_umID": int,
        "rbe3_wcgID": int,
    },
    "SPC1": {"C": int, "SID": int, "spc1_gridsetID": int},
}
